<?php

if ( !defined('ABSPATH') ) exit;

class QuickCab_Geofence_Rule {

  public $ID = 0;
  public $post_author = 0;
  public $post_date = '0000-00-00 00:00:00';
  public $post_date_gmt = '0000-00-00 00:00:00';
  public $post_content = '';
  public $post_title = '';
  public $post_excerpt = '';
  public $post_status = 'publish';
  public $comment_status = 'open';
  public $ping_status = 'open';
  public $post_password = '';
  public $post_name = '';
  public $to_ping = '';
  public $pinged = '';
  public $post_modified = '0000-00-00 00:00:00';
  public $post_modified_gmt = '0000-00-00 00:00:00';
  public $post_content_filtered = '';
  public $post_parent = 0;
  public $guid = '';
  public $menu_order = 0;
  public $post_mime_type = '';
  public $comment_count = 0;
  public $filter;

  private $starting_coords;
  private $destination_coords;
  private $pricing_structure;
  private $fixed_price_amount;
  private $increase_price_amount;
  private $vehicle_ids;
  private $direction;
  private $starting_polygon_center;
  private $starting_polygon_zoom;
  private $destination_polygon_center;
  private $destination_polygon_zoom;

  private $wp_post;
  private $meta_prefix;

  public function __construct($id) {
    $this->meta_prefix = '_quickcab_geofence_';
    $geofence_rule = WP_Post::get_instance($id);

    if ( $geofence_rule ) {
      $this->setup_geofence_rule($geofence_rule);
    } else {
      return false;
    }
  }

  public static function find_all() {
    $pricing_conditions = array();

    $post_search_args = array(
      'post_type' => 'geofence_rule',
      'post_staus' => array('publish'),
      'posts_per_page' => '-1'
    );
    $post_search = new WP_Query($post_search_args);

    foreach ($post_search->posts as $pricing_condition) {
      $unacceptable_status = array('auto-draft', 'future', 'draft', 'private', 'trash');

      if (
        in_array($pricing_condition->post_status, $unacceptable_status) ||
        get_post_status($pricing_condition->ID) == 'trash'// trashed
      ) {
        continue;
      }

      array_push($pricing_conditions, new QuickCab_Geofence_Rule($pricing_condition->ID));
    }

    return $pricing_conditions;
  }

  /**
   * Return the ID of the WP_Post object
   *
   * @since  1.0
   * @return int  order ID
   */
  public function get_ID() {
    return $this->ID;
  }

  /**
   * Return the coordinates of the first polygon for this geofence rule
   *
   * @since  1.0
   * @return array  the array of starting coordinates
   */
  public function get_starting_coords() {
    $starting_coords = quickcab_get_coords_from_string($this->starting_coords);
    return apply_filters('quickcab_geofence_rule_get_starting_coords', $starting_coords, $this->get_ID());
  }

  /**
   * Return the coordinates of the first polygon for this geofence rule in
   * string format
   *
   * @since  1.0
   * @return string  the starting coordinates
   */
  public function get_starting_coords_string() {
    return apply_filters('quickcab_geofence_rule_get_starting_coords_string', $this->starting_coords, $this->get_ID());
  }

  /**
   * Return the coordinates of the second polygon for this geofence rule
   *
   * @since  1.0
   * @return array  the array of destination coordinates
   */
  public function get_destination_coords() {
    $destination_coords = quickcab_get_coords_from_string($this->destination_coords);
    return apply_filters('quickcab_geofence_rule_get_destination_coords', $destination_coords, $this->get_ID());
  }

  /**
   * Return the coordinates of the second polygon for this geofence rule in
   * string format
   *
   * @since  1.0
   * @return string  the destination coordinates
   */
  public function get_destination_coords_string() {
    return apply_filters('quickcab_geofence_rule_get_destination_coords_string', $this->destination_coords, $this->get_ID());
  }

  /**
   * Return the pricing structure of the geofence rule
   *
   * @since  1.0
   * @return string  the pricing structure of the geofence rule (fixed/increase)
   */
  public function get_pricing_structure() {
    return apply_filters('quickcab_geofence_rule_get_pricing_structure', $this->pricing_structure, $this->get_ID());
  }

  /**
   * Is the pricing structure fixed price?
   *
   * @since   1.0
   * @return  boolean    fixed price?
   */
  public function is_fixed_price() {
    return $this->get_pricing_structure() == 'fixed-price';
  }

  /**
   * Is the pricing structure increased price?
   *
   * @since   1.0
   * @return  boolean    increased price?
   */
  public function is_increased_price() {
    return $this->get_pricing_structure() == 'increased-price';
  }

  /**
   * Return the fixed price amount of the geofence rule
   *
   * @since  1.0
   * @return float  the fixed price amount of the geofence rule
   */
  public function get_fixed_price_amount() {
    return apply_filters('quickcab_geofence_rule_get_fixed_price_amount', $this->fixed_price_amount, $this->get_ID());
  }

  /**
   * Return the increase price amount of the geofence rule
   *
   * @since  1.0
   * @return float  the increase price amount of the geofence rule
   */
  public function get_increase_price_amount() {
    return apply_filters('quickcab_geofence_rule_get_increase_price_amount', $this->increase_price_amount, $this->get_ID());
  }

  /**
   * Return the ids of all vehicles this rule applies to
   *
   * @since  1.0
   * @return array  the ids of all vehicles this rule applies to
   */
  public function get_vehicle_ids() {
    return apply_filters('quickcab_geofence_rule_get_vehicle_id', $this->vehicle_ids, $this->get_ID());
  }

  /**
   * Return the increase price amount of the geofence rule
   *
   * @since  1.0
   * @return QuickCab_Vehicle  the vehicles associated with this rule
   */
  public function get_vehicles() {
    $vehicles = array();

    foreach ( $this->get_vehicle_ids() as $vehicle_id ) {
      array_push($vehicles, new QuickCab_Vehicle($vehicle_id));
    }

    return $vehicles;
  }

  /**
   * Return the direction of the geofence rule
   *
   * @since  1.0
   * @return string  the direction of the geofence rule
   */
  public function get_direction() {
    return apply_filters('quickcab_geofence_rule_get_direction', $this->direction, $this->get_ID());
  }

  /**
   * Does the geofence rule only apply in one direction?
   *
   * @since   1.0
   * @return  boolean    only one direction?
   */
  public function is_one_way() {
    return $this->get_direction() == 'one-way';
  }

  /**
   * Does the geofence rule apply in both directions?
   *
   * @since   1.0
   * @return  boolean    both directions?
   */
  public function is_both_ways() {
    return $this->get_direction() == 'both-ways';
  }

  /**
   * Return the center of the starting polygon for the geofence rule
   *
   * @since  1.0
   * @return string  the center of the starting polygon for the geofence rule
   */
  public function get_starting_polygon_center() {
    return apply_filters('quickcab_geofence_rule_get_starting_polygon_center', $this->starting_polygon_center, $this->get_ID());
  }

  /**
   * Return the zoom of the starting polygon for the geofence rule
   *
   * @since  1.0
   * @return string  the zoom of the starting polygon for the geofence rule
   */
  public function get_starting_polygon_zoom() {
    return apply_filters('quickcab_geofence_rule_get_starting_polygon_zoom', $this->starting_polygon_zoom, $this->get_ID());
  }

  /**
   * Return the center of the destination polygon for the geofence rule
   *
   * @since  1.0
   * @return string  the center of the destination polygon for the geofence rule
   */
  public function get_destination_polygon_center() {
    return apply_filters('quickcab_geofence_rule_get_destination_polygon_center', $this->destination_polygon_center, $this->get_ID());
  }

  /**
   * Return the zoom of the destination polygon for the geofence rule
   *
   * @since  1.0
   * @return string  the zoom of the destination polygon for the geofence rule
   */
  public function get_destination_polygon_zoom() {
    return apply_filters('quickcab_geofence_rule_get_destination_polygon_zoom', $this->destination_polygon_zoom, $this->get_ID());
  }

  /**
   * Use the map or anywhere for the starting location
   *
   * @since  1.0
   * @return string  map or anywhere for starting location
   */
  public function get_starting_map_anywhere() {
    return apply_filters('quickcab_geofence_rule_get_starting_map_anywhere', $this->starting_map_anywhere, $this->get_ID());
  }

  /**
   * Is the starting area anywhere
   *
   * @since  1.0
   * @return boolean
   */
  public function starting_area_anywhere() {
    return $this->get_starting_map_anywhere() == 'anywhere';
  }

  /**
   * Is the starting area map
   *
   * @since  1.0
   * @return boolean
   */
  public function starting_area_map() {
    return $this->get_starting_map_anywhere() == 'map';
  }

  /**
   * Use the map or anywhere for the starting location
   *
   * @since  1.0
   * @return string  map or anywhere for starting location
   */
  public function get_destination_map_anywhere() {
    return apply_filters('quickcab_geofence_rule_get_destination_map_anywhere', $this->destination_map_anywhere, $this->get_ID());
  }

  /**
   * Is the destination area anywhere?
   *
   * @since  1.0
   * @return boolean
   */
  public function destination_area_anywhere() {
    return $this->get_destination_map_anywhere() == 'anywhere';
  }

  /**
   * Is the destination area map
   *
   * @since  1.0
   * @return boolean
   */
  public function destination_area_map() {
    return $this->get_destination_map_anywhere() == 'map';
  }

  /**
   * Should this rule override any surge pricing rules that would increase the price of the fare?
   *
   * @since  1.0
   * @return boolean
   */
  public function ignore_surge_pricing_rule() {
    return apply_filters('quickcab_geofence_rule_ignore_surge_pricing_rule', $this->ignore_surge_pricing_rule, $this->get_ID());
  }

  /**
   * Get the meta data for the geofence rule
   *
   * @since  1.0
   * @param  string  $key     the key to fetch
   * @param  boolean $single  get single data or array
   * @return [array|string]   value of meta data if single = true, array if not
   */
  public function get_meta($key = '', $single = true) {
    $value = get_post_meta($this->get_ID(), $this->meta_prefix . $key, $single);
    return apply_filters('quickcab_geofence_rule_get_meta_' . $key, $value, $this->get_ID(), $single);
  }

  /**
   * Update the meta data for the geofence rule
   *
   * @since  1.0
   * @param  string $key         the key to update
   * @param  string $value       the new value
   * @param  string $prev_value  the old value
   * @return [int|bool]          meta ID if key didn't exist, true on success, false on fail
   */
  public function update_meta($key = '', $value = '', $prev_value = '') {
    if ( empty($key) || $key == '' ) {
      return false;
    }

    $key = $this->meta_prefix . $key;

    $value = apply_filters('quickcab_geofence_rule_update_meta_' . $key, $value, $this->get_ID());
    return update_post_meta($this->get_ID(), $key, $value, $prev_value);
  }

  /**
   * Delete the meta data for the geofence rule
   *
   * @since  1.0
   * @param  string $key  the key to delete
   * @return bool         success or not
   */
  public function delete_meta($key) {
    if ( empty($key) || $key == '' ) {
      return false;
    }

    $key = $this->meta_prefix . $key;

    return delete_post_meta($this->get_ID(), $key);
  }

  /**
   * Does the current rule apply to the given ID
   *
   * @since  1.0
   * @param  int  $id  ID of the vehicle
   * @return bool      does the rule apply?
   */
  public function rule_applies_to($id) {
    return in_array(0, $this->get_vehicle_ids()) || in_array($id, $this->get_vehicle_ids());
  }

  /**
   * Get the image path for the starting polygon static map
   *
   * @since  1.0
   * @return string  the URL of the starting polygon image static map
   */
  public function get_starting_map_image_path() {
    return $this->get_map_image_path('starting');
  }

  /**
   * Get the image path for the destination polygon static map
   *
   * @since  1.0
   * @return string  the URL of the destination polygon image static map
   */
  public function get_destination_map_image_path() {
    return $this->get_map_image_path('destination');
  }

  private function get_map_image_path($type = null) {
    if ( is_null($type) ) {
      return null;
    }

    $images = get_attached_media('image', $this->get_ID());

    foreach ( $images as $image ) {
      if ( $image->post_title == $type . '-geofence-map' ) {
        return wp_get_attachment_image_src($image->ID)[0];
      }
    }

    return null;
  }

  public function get_starting_polygon_maps_image_url($size = 300) {
    return $this->get_polygon_maps_image_url(
      $this->get_starting_polygon_center(),
      $this->get_starting_polygon_zoom(),
      $size,
      $this->get_static_maps_starting_coords()
    );
  }

  public function get_destination_polygon_maps_image_url($size = 300) {
    return $this->get_polygon_maps_image_url(
      $this->get_destination_polygon_center(),
      $this->get_destination_polygon_zoom(),
      $size,
      $this->get_static_maps_destination_coords()
    );
  }

  /**
   * Get the google static map URL for the given settings
   *
   * @since  1.0
   * @param  string $centre  the centre of the map
   * @param  int    $zoom    the zoom of the map
   * @param  int    $size    the size of the map (size x size)
   * @param  string $coords  the coordinates of the polygon
   * @return string          the URL of the map
   */
  private function get_polygon_maps_image_url($centre, $zoom, $size, $coords) {
    $api_key = quickcab_get_option('google_static_maps_api_key');

    return "http://maps.googleapis.com/maps/api/staticmap?center=" . $centre .
      "&zoom=" . $zoom . "&size=" . $size . "x" . $size .
      "&sensor=false&path=color:blue|weight:3|fillcolor:blue|" . $coords .
      "&format=jpg&key=" . $api_key;
  }

  /**
   * Get coords of starting shape and convert them to string format
   *
   * @since  1.0
   * @return string  the coords in format: lat,long|lat,long|lat,long...
   */
  private function get_static_maps_starting_coords() {
    return $this->serialise_polygon_coords($this->get_starting_coords());
  }

  /**
   * Get coords of destination shape and convert them to string format
   *
   * @since  1.0
   * @return string  the coords in format: lat,long|lat,long|lat,long...
   */
  private function get_static_maps_destination_coords() {
    return $this->serialise_polygon_coords($this->get_destination_coords());
  }

  /**
   * Gets coordinates in array format, repeats the first coordinate and
   * serialises the output
   *
   * @since  1.0
   * @param  array  $coords  the array of coordinates
   * @return string          the serialised coords
   */
  private function serialise_polygon_coords($coords) {
    // Starting point is used twice to ensure to polygon is connected
    $gmaps_coords = array_merge($coords, array($coords[0]));

    foreach ( $gmaps_coords as $key => $gmaps_coord ) {
      $gmaps_coords[$key] = implode(',', array_map('strval', $gmaps_coord));
    }

    return implode('|', $gmaps_coords);
  }

  /**
   * Populates the default WordPress Post fields to QuickCab_Geofence_Rule
   *
   * @since  1.0
   * @param  WP_Post $post  the WP_Post
   * @return bool           if the save was successful
   */
  private function populate_post_fields($post) {
    if ( !is_object($post) || !is_a($post, 'WP_Post') || $post->post_type !== 'geofence_rule' ) {
      return false;
    }

    foreach ( $post as $key => $value ) {
      $this->$key = $value;
    }

    return true;
  }

  /**
   * Setup the object variables
   *
   * @since  1.0
   * @param  WP_Post $geofence_rule  the geofence rule WP_Post
   * @return bool                    did the setup suceed?
   */
  private function setup_geofence_rule($geofence_rule) {
    if ( is_null($geofence_rule) || !is_a($geofence_rule, 'WP_Post') || $geofence_rule->post_type !== 'geofence_rule' ) {
      return false;
    }

    do_action('quickcab_before_setup_geofence_rule', $this, $geofence_rule);

    $this->ID = $geofence_rule->ID;
    $this->wp_post = $geofence_rule;

    $this->starting_coords             = $this->setup_starting_coords();
    $this->destination_coords          = $this->setup_destination_coords();
    $this->pricing_structure           = $this->setup_pricing_structure();
    $this->fixed_price_amount          = $this->setup_fixed_price_amount();
    $this->increase_price_amount       = $this->setup_increase_price_amount();
    $this->vehicle_ids                 = $this->setup_vehicle_ids();
    $this->direction                   = $this->setup_direction();
    $this->starting_polygon_center     = $this->setup_starting_polygon_center();
    $this->starting_polygon_zoom       = $this->setup_starting_polygon_zoom();
    $this->destination_polygon_center  = $this->setup_destination_polygon_center();
    $this->destination_polygon_zoom    = $this->setup_destination_polygon_zoom();
    $this->starting_map_anywhere       = $this->setup_starting_map_anywhere();
    $this->destination_map_anywhere    = $this->setup_destination_map_anywhere();
    $this->ignore_surge_pricing_rule = $this->setup_ignore_surge_pricing_rule();

    $this->populate_post_fields($this->wp_post);

    do_action('quickcab_after_setup_geofence_rule', $this, $geofence_rule);

    return true;
  }

  private function setup_starting_coords() {
    return $this->get_meta('starting_coords');
  }

  private function setup_destination_coords() {
    return $this->get_meta('destination_coords');
  }

  private function setup_pricing_structure() {
    return $this->get_meta('pricing_structure');
  }

  private function setup_fixed_price_amount() {
    return (float) $this->get_meta('fixed_price_amount');
  }

  private function setup_increase_price_amount() {
    return (float) $this->get_meta('increase_price_amount');
  }

  private function setup_vehicle_ids() {
    if ( empty($this->get_meta('vehicle_ids')) ) {
      return array();
    }

    return array_map('intval', explode(',', $this->get_meta('vehicle_ids')));
  }

  private function setup_direction() {
    return $this->get_meta('direction');
  }

  private function setup_starting_polygon_center() {
    return $this->get_meta('starting_polygon_center');
  }

  private function setup_starting_polygon_zoom() {
    return (int) $this->get_meta('starting_polygon_zoom');
  }

  private function setup_destination_polygon_center() {
    return $this->get_meta('destination_polygon_center');
  }

  private function setup_destination_polygon_zoom() {
    return (int) $this->get_meta('destination_polygon_zoom');
  }

  private function setup_starting_map_anywhere() {
    return $this->get_meta('starting_map_anywhere');
  }

  private function setup_destination_map_anywhere() {
    return $this->get_meta('destination_map_anywhere');
  }

  private function setup_ignore_surge_pricing_rule() {
    return (bool) $this->get_meta('ignore_surge_pricing_rule');
  }
}
