<?php

if ( !defined('ABSPATH') ) exit;

class QuickCab_Surge_Price_Rule {

  public $ID = 0;
  public $post_author = 0;
  public $post_date = '0000-00-00 00:00:00';
  public $post_date_gmt = '0000-00-00 00:00:00';
  public $post_content = '';
  public $post_title = '';
  public $post_excerpt = '';
  public $post_status = 'publish';
  public $comment_status = 'open';
  public $ping_status = 'open';
  public $post_password = '';
  public $post_name = '';
  public $to_ping = '';
  public $pinged = '';
  public $post_modified = '0000-00-00 00:00:00';
  public $post_modified_gmt = '0000-00-00 00:00:00';
  public $post_content_filtered = '';
  public $post_parent = 0;
  public $guid = '';
  public $menu_order = 0;
  public $post_mime_type = '';
  public $comment_count = 0;
  public $filter;

  private $date_structure;
  private $fixed_start_date;
  private $fixed_start_time_hour;
  private $fixed_start_time_minute;
  private $fixed_end_date;
  private $fixed_end_time_hour;
  private $fixed_end_time_minute;
  private $day_tiers;
  private $pricing_structure;
  private $increase_amount;
  private $increase_multiplier;

  private $wp_post;
  private $meta_prefix;

  public function __construct($id) {
    $this->meta_prefix = '_quickcab_surge_price_';
    $surge_price_rule = WP_Post::get_instance($id);

    if ( $surge_price_rule ) {
      $this->setup_surge_price_rule($surge_price_rule);
    } else {
      return false;
    }
  }

  /**
   * Return the ID of the WP_Post object
   *
   * @since  1.0
   * @return int  order ID
   */
  public function get_ID() {
    return $this->ID;
  }

  /**
   * Return the date structure of this surge price rule
   *
   * @since  1.0
   * @return string  the date structure
   */
  public function get_date_structure() {
    return apply_filters('quickcab_surge_price_rule_get_date_structure', $this->date_structure, $this->get_ID());
  }

  /**
   * Return the fixed start time hour of this surge price rule
   *
   * @since  1.0
   * @return string  the fixed start time hour
   */
  public function get_fixed_start_time_hour() {
    return apply_filters('quickcab_surge_price_rule_get_fixed_start_time_hour', sprintf('%02d', $this->fixed_start_time_hour), $this->get_ID());
  }

  /**
   * Return the fixed start time minute of this surge price rule
   *
   * @since  1.0
   * @return string  the fixed start time minute
   */
  public function get_fixed_start_time_minute() {
    return apply_filters('quickcab_surge_price_rule_get_fixed_start_time_minute', sprintf('%02d', $this->fixed_start_time_minute), $this->get_ID());
  }

  /**
   * Return the fixed start date of this surge price rule
   *
   * @since  1.0
   * @return string  the fixed start date
   */
  public function get_fixed_start_date() {
    return apply_filters('quickcab_surge_price_rule_get_start_date', $this->fixed_start_date, $this->get_ID());
  }

  /**
   * Return the start date of this surge price rule
   *
   * @since  1.0
   * @return int  the start date (timestamp)
   */
  public function get_start_date_timestamp() {
    $start_date_str = $this->get_fixed_start_date() . ' ' . $this->get_fixed_start_time_hour() . ':' . $this->get_fixed_start_time_minute();
    return DateTime::createFromFormat("d/m/Y H:i", $start_date_str)->format('U');
  }

  /**
   * Return the fixed end time hour of this surge price rule
   *
   * @since  1.0
   * @return string  the fixed end time hour
   */
  public function get_fixed_end_time_hour() {
    return apply_filters('quickcab_surge_price_rule_get_fixed_end_time_hour', sprintf('%02d', $this->fixed_end_time_hour), $this->get_ID());
  }

  /**
   * Return the fixed end time minute of this surge price rule
   *
   * @since  1.0
   * @return string  the fixed end time minute
   */
  public function get_fixed_end_time_minute() {
    return apply_filters('quickcab_surge_price_rule_get_fixed_end_time_minute', sprintf('%02d', $this->fixed_end_time_minute), $this->get_ID());
  }

  /**
   * Return the fixed end date of this surge price rule
   *
   * @since  1.0
   * @return string  the fixed end date
   */
  public function get_fixed_end_date() {
    return apply_filters('quickcab_surge_price_rule_get_fixed_end_date', $this->fixed_end_date, $this->get_ID());
  }

  /**
   * Return the end date of this surge price rule
   *
   * @since  1.0
   * @return string  the end date
   */
  public function get_end_date_timestamp() {
    $end_date_str = $this->get_fixed_end_date() . ' ' . $this->get_fixed_end_time_hour() . ':' . $this->get_fixed_end_time_minute();
    return DateTime::createFromFormat("d/m/Y H:i", $end_date_str)->format('U');
  }

  /**
   * Return the day tiers of this surge price rule
   *
   * @since  1.0
   * @return string  the day tiers
   */
  public function get_day_tiers() {
    return apply_filters('quickcab_surge_price_rule_get_day_tiers', $this->day_tiers, $this->get_ID());
  }

  /**
   * Return the pricing structure of this surge price rule
   *
   * @since  1.0
   * @return string  the pricing structure
   */
  public function get_pricing_structure() {
    return apply_filters('quickcab_surge_price_rule_get_pricing_structure', $this->pricing_structure, $this->get_ID());
  }

  /**
   * Return the increase amount of this surge price rule
   *
   * @since  1.0
   * @return string  the increase amount
   */
  public function get_increase_amount() {
    return apply_filters('quickcab_surge_price_rule_get_increase_amount', $this->increase_amount, $this->get_ID());
  }

  /**
   * Return the increase multiplier of this surge price rule
   *
   * @since  1.0
   * @return string  the increase multiplier
   */
  public function get_increase_multiplier() {
    return apply_filters('quickcab_surge_price_rule_get_increase_multiplier', $this->increase_multiplier, $this->get_ID());
  }

  /**
   * Does the surge price rule increase by a fixed amount?
   *
   * @since  1.0
   * @return boolean
   */
  public function is_increase_amount() {
    return $this->get_pricing_structure() == 'increase-amount';
  }

  /**
   * Does the surge price rule increase by a multiplier?
   *
   * @since  1.0
   * @return bool
   */
  public function is_increase_multiplier() {
    return $this->get_pricing_structure() == 'increase-multiplier';
  }

  /**
   * Does the surge price rule use fixed dates?
   *
   * @since  1.0
   * @return bool
   */
  public function uses_fixed_dates() {
    return $this->get_date_structure() == 'fixed-date';
  }

  /**
   * Does the surge price rule use days of the week?
   *
   * @since   1.0
   * @return  bool
   */
  public function uses_days_of_week() {
    return $this->get_date_structure() == 'days-of-week';
  }

  /**
   * Does the surge price rule apply to the given journey?
   *
   * @since  1.0
   * @param  QuickCab_Journey  $journey  the journey to check
   * @return bool
   */
  public function applies_to($journey) {
    if ( $this->uses_fixed_dates() ) {
      return $this->applies_to_fixed_date($journey);
    }

    if ( $this->uses_days_of_week() ) {
      return $this->applies_to_day_tiers($journey);
    }

    return false;
  }

  private function applies_to_fixed_date($journey) {
    $surge_rule_start = $this->get_start_date_timestamp();
    $surge_rule_end = $this->get_end_date_timestamp();

    $journey_start = $journey->get_departure_time();
    $journey_end = $journey_start + $journey->get_duration_in_seconds();

    // do the 2 ranges overlap?
    return $surge_rule_start <= $journey_end && $surge_rule_end >= $journey_start;
  }

  private function applies_to_day_tiers($journey) {
    $journey_start = $journey->get_departure_time();
    $journey_end = $journey_start + $journey->get_duration_in_seconds();
    $journey_start_day = intval(date('w', $journey_start));
    $journey_end_day = intval(date('w', $journey_end));

    foreach ( $this->get_day_tiers() as $day_tier ) {
      // the number of times midnight is passed on the tier
      $tier_day_span = $day_tier['to_day'] + ($day_tier['from_day'] > $day_tier['to_day'] ? 7: 0) - $day_tier['from_day'];

      for ( $i = 0; $i <= $tier_day_span; $i++ ) {
        // current day of week (will be 0-6 inclusive)
        $current_tier_day = $day_tier['from_day'] + $i;

        // keep $current_tier_day in range 0-6
        if ( $current_tier_day > 6 ) {
          $current_tier_day = $current_tier_day - 7;
        }

        if ( quickcab_day_in_day_range($current_tier_day, $journey_start_day, $journey_end_day) ) {
          $mod_journey_end_day = $journey_end_day + ($journey_end_day < $journey_start_day ? 7 : 0);

          // minutes since midnight
          $tier_start_mins = $day_tier['from_hour'] * 60 + $day_tier['from_min'];
          $tier_end_mins = $day_tier['to_hour'] * 60 + $day_tier['to_min'];

          // minutes since midnight
          $journey_start_mins = quickcab_mins_since_midnight($journey_start);
          $journey_end_mins = quickcab_mins_since_midnight($journey_end);

          // if were not on the last day of the tier
          if ( $i < $tier_day_span ) {
            $tier_end_mins = 24 * 60;
          }

          // if were not on the first day of the tier
          if ( $i > 0 ) {
            $tier_start_mins = 0;
          }

          // if were not on the first day of the journey
          if ( $day_tier['from_day'] + $i < $mod_journey_end_day ) {
            $journey_end_mins = 24 * 60;
          }

          if ( $current_tier_day > $journey_start_day ) {
            $journey_start_mins = 0;
          }

          // we dont want to return if this is false as other tiers may be suitable
          if ( $tier_start_mins <= $journey_end_mins && $tier_end_mins >= $journey_start_mins ) {
            return true;
          }
        }
      }
    }

    return false;
  }

  /**
   * Gets all QuickCab_Surge_Price_Rule
   *
   * @since  1.0
   * @return array  array of all surge price rules
   */
  public static function find_all() {
    $surge_price_rules = array();

    $post_search = new WP_Query(array(
      'post_type' => 'surge_price_rule',
      'post_staus' => array('publish'),
      'posts_per_page' => -1
    ));

    $unacceptable_status = apply_filters('quickcab_surge_price_rule_unacceptable_status', array(
      'auto-draft', 'future', 'draft', 'private', 'trash'
    ));

    foreach ( $post_search->posts as $surge_price_rule ) {
      if ( in_array($surge_price_rule->post_status, $unacceptable_status) ) {
        continue;
      }

      $surge_price_rules[$surge_price_rule->ID] = new QuickCab_Surge_Price_Rule($surge_price_rule->ID);
    }

    return $surge_price_rules;
  }

  /**
   * Get the meta data for the geofence rule
   *
   * @since  1.0
   * @param  string  $key     the key to fetch
   * @param  boolean $single  get single data or array
   * @return [array|string]   value of meta data if single = true, array if not
   */
  public function get_meta($key = '', $single = true) {
    $value = get_post_meta($this->get_ID(), $this->meta_prefix . $key, $single);
    return apply_filters('quickcab_surge_price_rule_get_meta_' . $key, $value, $this->get_ID(), $single);
  }

  /**
   * Update the meta data for the geofence rule
   *
   * @since  1.0
   * @param  string $key         the key to update
   * @param  string $value       the new value
   * @param  string $prev_value  the old value
   * @return [int|bool]          meta ID if key didn't exist, true on success, false on fail
   */
  public function update_meta($key = '', $value = '', $prev_value = '') {
    if ( empty($key) || $key == '' ) {
      return false;
    }

    $key = $this->meta_prefix . $key;

    $value = apply_filters('quickcab_surge_price_rule_update_meta_' . $key, $value, $this->get_ID());
    return update_post_meta($this->get_ID(), $key, $value, $prev_value);
  }

  /**
   * Delete the meta data for the geofence rule
   *
   * @since  1.0
   * @param  string $key  the key to delete
   * @return bool         success or not
   */
  public function delete_meta($key) {
    if ( empty($key) || $key == '' ) {
      return false;
    }

    $key = $this->meta_prefix . $key;

    return delete_post_meta($this->get_ID(), $key);
  }

  /**
   * Populates the default WordPress Post fields to QuickCab_Geofence_Rule
   *
   * @since  1.0
   * @param  WP_Post $post  the WP_Post
   * @return bool           if the save was successful
   */
  private function populate_post_fields($post) {
    if ( !is_object($post) || !is_a($post, 'WP_Post') || $post->post_type !== 'geofence_rule' ) {
      return false;
    }

    foreach ( $post as $key => $value ) {
      $this->$key = $value;
    }

    return true;
  }

  /**
   * Setup the object variables
   *
   * @since  1.0
   * @param  WP_Post $surge_price_rule  the surge price rule WP_Post
   * @return bool                       did the setup suceed?
   */
  private function setup_surge_price_rule($surge_price_rule) {
    if ( is_null($surge_price_rule) || !is_a($surge_price_rule, 'WP_Post') || $surge_price_rule->post_type !== 'surge_price_rule' ) {
      return false;
    }

    do_action('quickcab_before_setup_surge_price_rule', $this, $surge_price_rule);

    $this->ID      = $surge_price_rule->ID;
    $this->wp_post = $surge_price_rule;

    $this->date_structure          = $this->setup_date_structure();
    $this->fixed_start_time_hour   = $this->setup_fixed_start_time_hour();
    $this->fixed_start_time_minute = $this->setup_fixed_start_time_minute();
    $this->fixed_start_date        = $this->setup_fixed_start_date();
    $this->fixed_end_time_hour     = $this->setup_fixed_end_time_hour();
    $this->fixed_end_time_minute   = $this->setup_fixed_end_time_minute();
    $this->fixed_end_date          = $this->setup_fixed_end_date();
    $this->day_tiers               = $this->setup_day_tiers();
    $this->pricing_structure       = $this->setup_pricing_structure();
    $this->increase_amount         = $this->setup_increase_amount();
    $this->increase_multiplier     = $this->setup_increase_multiplier();

    $this->populate_post_fields($this->wp_post);

    do_action('quickcab_after_setup_surge_price_rule', $this, $surge_price_rule);

    return true;
  }

  private function setup_date_structure() {
    return $this->get_meta('date_structure');
  }

  private function setup_fixed_start_time_hour() {
    return $this->get_meta('fixed_start_time_hour');
  }
  private function setup_fixed_start_time_minute() {
    return $this->get_meta('fixed_start_time_minute');
  }

  private function setup_fixed_start_date() {
    return $this->get_meta('fixed_start_date');
  }

  private function setup_fixed_end_time_hour() {
    return $this->get_meta('fixed_end_time_hour');
  }
  private function setup_fixed_end_time_minute() {
    return $this->get_meta('fixed_end_time_minute');
  }

  private function setup_fixed_end_date() {
    return $this->get_meta('fixed_end_date');
  }

  private function setup_day_tiers() {
    $tiers = unserialize($this->get_meta('day_tiers'));

    if ( !$tiers ) {
      $tiers = array();
    }

    foreach ( $tiers as $key => $tier ) {
      foreach ( $tier as $k => $v ) {
        $tier[$k] = intval($v);
      }

      $tiers[$key] = $tier;
    }

    return $tiers;
  }

  private function setup_pricing_structure() {
    return $this->get_meta('pricing_structure');
  }

  private function setup_increase_amount() {
    return (float) $this->get_meta('increase_amount');
  }

  private function setup_increase_multiplier() {
    return (float) $this->get_meta('increase_multiplier');
  }
}
